# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides interface to G90 alarm panel.

.. note:: Only protocol 1.2 is supported!

The next example queries the device with IP address `10.10.10.250` for the
information - the product name, protocol version, HW versions and such.

>>> import asyncio
>>> from pyg90alarm import G90Alarm
>>> async def main():
       g90 = G90Alarm(host='10.10.10.250')
       s = await g90.host_info
       print(s)
>>> asyncio.run(main())
G90HostInfo(host_guid='<...>',
            product_name='TSV018-C3SIA',
            wifi_protocol_version='1.2',
            cloud_protocol_version='1.1',
            mcu_hw_version='206',
            wifi_hw_version='206',
            gsm_status=3,
            wifi_status=3,
            reserved1=0,
            reserved2=0,
            band_frequency='4264',
            gsm_signal_level=50,
            wifi_signal_level=100)

"""
from __future__ import annotations
import asyncio
from asyncio import Task
import logging
from typing import (
    TYPE_CHECKING, Any, List, Optional, AsyncGenerator,
    Callable, Coroutine, Union
)
from .const import (
    G90Commands, REMOTE_PORT,
    REMOTE_TARGETED_DISCOVERY_PORT,
    LOCAL_TARGETED_DISCOVERY_PORT,
    LOCAL_NOTIFICATIONS_HOST,
    LOCAL_NOTIFICATIONS_PORT,
    G90ArmDisarmTypes,
    G90RemoteButtonStates,
)
from .base_cmd import (G90BaseCommand, G90BaseCommandData)
from .paginated_result import G90PaginatedResult, G90PaginatedResponse
from .entities.sensor import (G90Sensor, G90SensorTypes)
from .entities.device import G90Device
from .device_notifications import (
    G90DeviceNotifications,
)
from .discovery import G90Discovery, G90DiscoveredDevice
from .targeted_discovery import (
    G90TargetedDiscovery, G90DiscoveredDeviceTargeted,
)
from .host_info import G90HostInfo
from .host_status import G90HostStatus
from .config import (G90AlertConfig, G90AlertConfigFlags)
from .history import G90History
from .user_data_crc import G90UserDataCRC
from .callback import G90Callback
from .exceptions import G90Error, G90TimeoutError

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    # Type alias for the callback functions available to the user, should be
    # compatible with `G90Callback.Callback` type, since `G90Callback.invoke`
    # is used to invoke them
    AlarmCallback = Union[
        Callable[[int, str, Any], None],
        Callable[[int, str, Any], Coroutine[None, None, None]]
    ]
    DoorOpenCloseCallback = Union[
        Callable[[int, str, bool], None],
        Callable[[int, str, bool], Coroutine[None, None, None]]
    ]
    SensorCallback = Union[
        Callable[[int, str, bool], None],
        Callable[[int, str, bool], Coroutine[None, None, None]]
    ]
    LowBatteryCallback = Union[
        Callable[[int, str], None],
        Callable[[int, str], Coroutine[None, None, None]]
    ]
    ArmDisarmCallback = Union[
        Callable[[G90ArmDisarmTypes], None],
        Callable[[G90ArmDisarmTypes], Coroutine[None, None, None]]
    ]
    SosCallback = Union[
        Callable[[int, str, bool], None],
        Callable[[int, str, bool], Coroutine[None, None, None]]
    ]
    RemoteButtonPressCallback = Union[
        Callable[[int, str, G90RemoteButtonStates], None],
        Callable[
            [int, str, G90RemoteButtonStates], Coroutine[None, None, None]
        ]
    ]
    # Sensor-related callbacks for `G90Sensor` class - despite that class
    # stores them, the invocation is done by the `G90Alarm` class hence these
    # are defined here
    SensorStateCallback = Union[
        Callable[[bool], None],
        Callable[[bool], Coroutine[None, None, None]]
    ]
    SensorLowBatteryCallback = Union[
        Callable[[], None],
        Callable[[], Coroutine[None, None, None]]
    ]


# pylint: disable=too-many-public-methods
class G90Alarm(G90DeviceNotifications):

    """
    Allows to interact with G90 alarm panel.

    :param host: Hostname or IP address of the alarm panel. Since the
     protocol is UDP-based it is ok to use broadcast or directed broadcast
     addresses, such as `255.255.255.255` or `10.10.10.255` (the latter assumes
     the device is on `10.10.10.0/24` network)
    :param port: The UDP port of the device where it listens for the
     protocol commands on WiFi interface, currently the devices don't allow it
     to be customized
    :param reset_occupancy_interval: The interval upon that the sensors are
     simulated to go into inactive state.
    """
    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(self, host: str, port: int = REMOTE_PORT,
                 reset_occupancy_interval: float = 3.0,
                 notifications_local_host: str = LOCAL_NOTIFICATIONS_HOST,
                 notifications_local_port: int = LOCAL_NOTIFICATIONS_PORT):
        super().__init__(
            local_host=notifications_local_host,
            local_port=notifications_local_port
        )
        self._host: str = host
        self._port: int = port
        self._sensors: List[G90Sensor] = []
        self._sensors_lock = asyncio.Lock()
        self._devices: List[G90Device] = []
        self._devices_lock = asyncio.Lock()
        self._notifications: Optional[G90DeviceNotifications] = None
        self._sensor_cb: Optional[SensorCallback] = None
        self._armdisarm_cb: Optional[ArmDisarmCallback] = None
        self._door_open_close_cb: Optional[DoorOpenCloseCallback] = None
        self._alarm_cb: Optional[AlarmCallback] = None
        self._low_battery_cb: Optional[LowBatteryCallback] = None
        self._sos_cb: Optional[SosCallback] = None
        self._remote_button_press_cb: Optional[
            RemoteButtonPressCallback
        ] = None
        self._reset_occupancy_interval = reset_occupancy_interval
        self._alert_config: Optional[G90AlertConfigFlags] = None
        self._sms_alert_when_armed = False
        self._alert_simulation_task: Optional[Task[Any]] = None
        self._alert_simulation_start_listener_back = False

    async def command(
        self, code: G90Commands, data: Optional[G90BaseCommandData] = None
    ) -> G90BaseCommandData:
        """
        Invokes a command against the alarm panel.

        :param code: Command code
        :param data: Command data
        :return: The result of command invocation
        """
        cmd: G90BaseCommand = await G90BaseCommand(
            self._host, self._port, code, data).process()
        return cmd.result

    def paginated_result(
        self, code: G90Commands, start: int = 1, end: Optional[int] = None
    ) -> AsyncGenerator[G90PaginatedResponse, None]:
        """
        Returns asynchronous generator for a paginated command, that is -
        command operating on a range of records.

        :param code: Command code
        :param start: Starting record position (one-based)
        :param end: Ending record position (one-based)
        :return: Asynchronous generator over the result of command invocation.
        """
        return G90PaginatedResult(
            self._host, self._port, code, start, end
        ).process()

    @classmethod
    async def discover(cls) -> List[G90DiscoveredDevice]:
        """
        Initiates discovering devices available in the same network segment, by
        using global broadcast address as the destination.

        :return: List of discovered devices
        """
        cmd: G90Discovery = await G90Discovery(
            port=REMOTE_PORT,
            host='255.255.255.255'
        ).process()
        return cmd.devices

    @classmethod
    async def targeted_discover(
        cls, device_id: str
    ) -> List[G90DiscoveredDeviceTargeted]:
        """
        Initiates discovering devices available in the same network segment
        using targeted protocol, that is - specifying target device GUID in the
        request, so only the specific device should respond to the query.

        :param device_id: GUID of the target device to discover
        :return: List of discovered devices
        """
        cmd = await G90TargetedDiscovery(
            device_id=device_id,
            port=REMOTE_TARGETED_DISCOVERY_PORT,
            local_port=LOCAL_TARGETED_DISCOVERY_PORT,
            host='255.255.255.255'
        ).process()
        return cmd.devices

    @property
    async def sensors(self) -> List[G90Sensor]:
        """
        Property over new :meth:`.get_sensors` method, retained for
        compatibility.
        """
        return await self.get_sensors()

    async def get_sensors(self) -> List[G90Sensor]:
        """
        Provides list of sensors configured in the device. Please note the list
        is cached upon first call, so you need to re-instantiate the class to
        reflect any updates there.

        :return: List of sensors
        """
        # Use lock around the operation, to ensure no duplicated entries in the
        # resulting list or redundant exchanges with panel are made when the
        # method is called concurrently
        async with self._sensors_lock:
            if not self._sensors:
                sensors = self.paginated_result(
                    G90Commands.GETSENSORLIST
                )
                async for sensor in sensors:
                    obj = G90Sensor(
                        *sensor.data, parent=self, subindex=0,
                        proto_idx=sensor.proto_idx
                    )
                    self._sensors.append(obj)

                _LOGGER.debug(
                    'Total number of sensors: %s', len(self._sensors)
                )

            return self._sensors

    async def find_sensor(self, idx: int, name: str) -> Optional[G90Sensor]:
        """
        Finds sensor by index and name.

        :param idx: Sensor index
        :param name: Sensor name
        :return: Sensor instance
        """
        sensors = await self.get_sensors()

        # Fast lookup by direct index
        if idx < len(sensors) and sensors[idx].name == name:
            sensor = sensors[idx]
            _LOGGER.debug('Found sensor via fast lookup: %s', sensor)
            return sensor

        # Fast lookup failed, perform slow one over the whole sensors list
        for sensor in sensors:
            if sensor.index == idx and sensor.name == name:
                _LOGGER.debug('Found sensor: %s', sensor)
                return sensor

        _LOGGER.error('Sensor not found: idx=%s, name=%s', idx, name)
        return None

    @property
    async def devices(self) -> List[G90Device]:
        """
        Property over new :meth:`.get_devices` method, retained for
        compatibility.
        """
        return await self.get_devices()

    async def get_devices(self) -> List[G90Device]:
        """
        Provides list of devices (switches) configured in the device. Please
        note the list is cached upon first call, so you need to re-instantiate
        the class to reflect any updates there. Multi-node devices, those
        having multiple ports, are expanded into corresponding number of
        resulting entries.

        :return: List of devices
        """
        # See `get_sensors` method for the rationale behind the lock usage
        async with self._devices_lock:
            if not self._devices:
                devices = self.paginated_result(
                    G90Commands.GETDEVICELIST
                )
                async for device in devices:
                    obj = G90Device(
                        *device.data, parent=self, subindex=0,
                        proto_idx=device.proto_idx
                    )
                    self._devices.append(obj)
                    # Multi-node devices (first node has already been added
                    # above
                    for node in range(1, obj.node_count):
                        obj = G90Device(
                            *device.data, parent=self,
                            subindex=node, proto_idx=device.proto_idx
                        )
                        self._devices.append(obj)

                _LOGGER.debug(
                    'Total number of devices: %s', len(self._devices)
                )

            return self._devices

    @property
    async def host_info(self) -> G90HostInfo:
        """
        Property over new :meth:`.get_host_info` method, retained for
        compatibility.
        """
        return await self.get_host_info()

    async def get_host_info(self) -> G90HostInfo:
        """
        Provides the device information (for example hardware versions, signal
        levels etc.).

        :return: Device information
        """
        res = await self.command(G90Commands.GETHOSTINFO)
        info = G90HostInfo(*res)
        self.device_id = info.host_guid
        return info

    @property
    async def host_status(self) -> G90HostStatus:
        """
        Property over new :meth:`.get_host_status` method, retained for
        compatibility.
        """
        return await self.get_host_status()

    async def get_host_status(self) -> G90HostStatus:
        """
        Provides the device status (for example, armed or disarmed, configured
        phone number, product name etc.).

        :return: Device information

        """
        res = await self.command(G90Commands.GETHOSTSTATUS)
        return G90HostStatus(*res)

    @property
    async def alert_config(self) -> G90AlertConfigFlags:
        """
        Property over new :meth:`.get_alert_config` method, retained for
        compatibility.
        """
        return await self.get_alert_config()

    async def get_alert_config(self) -> G90AlertConfigFlags:
        """
        Retrieves the alert configuration flags from the device. Please note
        the configuration is cached upon first call, so you need to
        re-instantiate the class to reflect any updates there.

        :return: The alerts configured
        """
        if not self._alert_config:
            self._alert_config = await self._alert_config_uncached()
        return self._alert_config

    async def _alert_config_uncached(self) -> G90AlertConfigFlags:
        """
        Retrieves the alert configuration flags directly from the device.

        :return: The alerts configured
        """
        res = await self.command(G90Commands.GETNOTICEFLAG)
        return G90AlertConfig(*res).flags

    async def set_alert_config(self, flags: G90AlertConfigFlags) -> None:
        """
        Sets the alert configuration flags on the device.
        """
        # Use uncached method retrieving the alert configuration, to ensure the
        # actual value retrieved from the device
        alert_config = await self._alert_config_uncached()
        if alert_config != self._alert_config:
            _LOGGER.warning(
                'Alert configuration changed externally,'
                ' overwriting (read "%s", will be set to "%s")',
                str(alert_config), str(flags)
            )
        await self.command(G90Commands.SETNOTICEFLAG, [flags.value])
        # Update the alert configuration stored
        self._alert_config = flags

    @property
    async def user_data_crc(self) -> G90UserDataCRC:
        """
        Property over new :meth:`.get_user_data_crc` method, retained for
        compatibility.
        """
        return await self.get_user_data_crc()

    async def get_user_data_crc(self) -> G90UserDataCRC:
        """
        Retieves checksums (CRC) for different on-device databases (history,
        sensors etc.). Might be used to detect if there is a change in a
        particular database.

        .. note:: Note that due to a bug in the firmware CRC for sensors and
          device databases change on each call even if there were no changes

        :return: Checksums for different databases
        """
        res = await self.command(G90Commands.GETUSERDATACRC)
        return G90UserDataCRC(*res)

    async def history(
        self, start: int = 1, count: int = 1
    ) -> List[G90History]:
        """
        Retrieves event history from the device.

        :param start: Starting record number (one-based)
        :param count: Number of records to retrieve
        :return: List of history entries
        """
        res = self.paginated_result(G90Commands.GETHISTORY,
                                    start, count)

        # Sort the history entries from older to newer - device typically does
        # that, but apparently that is not guaranteed
        return sorted(
            [G90History(*x.data) async for x in res],
            key=lambda x: x.datetime, reverse=True
        )

    async def on_sensor_activity(
        self, idx: int, name: str, occupancy: bool = True
    ) -> None:
        """
        Invoked both for sensor notifications and door open/close
        alerts, since the logic for both is same and could be reused.
        Fires corresponding callback if set by the user with
        :attr:`.sensor_callback`.

        Please note the method is for internal use by the class.

        :param idx: The index of the sensor the callback is invoked for.
         Please note the index is a property of sensor, not the direct index of
         :attr:`sensors` array
        :param name: The name of the sensor, along with the `idx` parameter
         it is used to look the sensor up from the :attr:`sensors` list
        :param occupancy: The flag indicating the target sensor state
         (=occupancy), will always be `True` for callbacks invoked from alarm
         panel notifications, and reflects actual sensor state for device
         alerts (only for `door` type sensors, if door open/close alerts are
         enabled)
        """
        _LOGGER.debug('on_sensor_activity: %s %s %s', idx, name, occupancy)
        sensor = await self.find_sensor(idx, name)
        if sensor:
            # Reset the low battery flag since the sensor reports activity,
            # implying it has sufficient battery power
            # pylint: disable=protected-access
            sensor._set_low_battery(False)
            # Set the sensor occupancy
            # pylint: disable=protected-access
            sensor._set_occupancy(occupancy)

            # Emulate turning off the occupancy - most of sensors will not
            # notify the device of that, nor the device would emit such
            # notification itself
            def reset_sensor_occupancy(sensor: G90Sensor) -> None:
                sensor._set_occupancy(False)
                G90Callback.invoke(sensor.state_callback, sensor.occupancy)

            # Determine if door close notifications are available for the given
            # sensor
            alert_config_flags = await self.alert_config
            door_close_alert_enabled = (
                G90AlertConfigFlags.DOOR_CLOSE in alert_config_flags)
            sensor_is_door = sensor.type == G90SensorTypes.DOOR

            # Alarm panel could emit door close alerts (if enabled) for sensors
            # of type `door`, and such event will be used to reset the
            # occupancy for the given sensor. Otherwise, the sensor closing
            # event will be emulated
            if not door_close_alert_enabled or not sensor_is_door:
                _LOGGER.debug("Sensor '%s' is not a door (type %s),"
                              ' or door close alert is disabled'
                              ' (alert config flags %s),'
                              ' closing event will be emulated upon'
                              ' %s seconds',
                              name, sensor.type, alert_config_flags,
                              self._reset_occupancy_interval)
                G90Callback.invoke_delayed(
                    self._reset_occupancy_interval,
                    reset_sensor_occupancy, sensor)

            # Invoke per-sensor callback if provided
            G90Callback.invoke(sensor.state_callback, occupancy)

        # Invoke global callback if provided
        G90Callback.invoke(self._sensor_cb, idx, name, occupancy)

    @property
    def sensor_callback(self) -> Optional[SensorCallback]:
        """
        Sensor activity callback, which is invoked when sensor activates.
        """
        return self._sensor_cb

    @sensor_callback.setter
    def sensor_callback(self, value: SensorCallback) -> None:
        self._sensor_cb = value

    async def on_door_open_close(
        self, event_id: int, zone_name: str, is_open: bool
    ) -> None:
        """
        Invoked when door open/close alert comes from the alarm
        panel. Fires corresponding callback if set by the user with
        :attr:`.door_open_close_callback`.

        Please note the method is for internal use by the class.

        .. seealso:: :meth:`.on_sensor_activity` method for arguments
        """
        # Same internal callback is reused both for door open/close alerts and
        # sensor notifications. The former adds reporting when a door is
        # closed, since the notifications aren't sent for such events
        await self.on_sensor_activity(event_id, zone_name, is_open)
        # Invoke user specified callback if any
        G90Callback.invoke(
            self._door_open_close_cb, event_id, zone_name, is_open
        )

    @property
    def door_open_close_callback(self) -> Optional[DoorOpenCloseCallback]:
        """
        The door open/close callback, which is invoked when door
        is opened or closed (if corresponding alert is configured on the
        device).
        """
        return self._door_open_close_cb

    @door_open_close_callback.setter
    def door_open_close_callback(self, value: DoorOpenCloseCallback) -> None:
        self._door_open_close_cb = value

    async def on_armdisarm(self, state: G90ArmDisarmTypes) -> None:
        """
        Invoked when the device is armed or disarmed.  Fires corresponding
        callback if set by the user with :attr:`.armdisarm_callback`.

        Please note the method is for internal use by the class.

        :param state: Device state (armed, disarmed, armed home)
        """
        if self._sms_alert_when_armed:
            if state == G90ArmDisarmTypes.DISARM:
                # Disable SMS alerts from the device
                await self.set_alert_config(
                    await self.alert_config & ~G90AlertConfigFlags.SMS_PUSH
                )
            if state in (G90ArmDisarmTypes.ARM_AWAY,
                         G90ArmDisarmTypes.ARM_HOME):
                # Enable SMS alerts from the device
                await self.set_alert_config(
                    await self.alert_config | G90AlertConfigFlags.SMS_PUSH
                )
        G90Callback.invoke(self._armdisarm_cb, state)

    @property
    def armdisarm_callback(self) -> Optional[ArmDisarmCallback]:
        """
        The device arm/disarm callback, which is invoked when device state
        changes.
        """
        return self._armdisarm_cb

    @armdisarm_callback.setter
    def armdisarm_callback(self, value: ArmDisarmCallback) -> None:
        self._armdisarm_cb = value

    async def on_alarm(self, event_id: int, zone_name: str) -> None:
        """
        Invoked when alarm is triggered. Fires corresponding callback if set by
        the user with :attr:`.alarm_callback`.

        Please note the method is for internal use by the class.

        :param event_id: Index of the sensor triggered alarm
        :param zone_name: Sensor name
        """
        sensor = await self.find_sensor(event_id, zone_name)
        # The callback is still delivered to the caller even if the sensor
        # isn't found, only `extra_data` is skipped. That is to ensure the
        # important callback isn't filtered
        extra_data = sensor.extra_data if sensor else None
        # Invoke the sensor activity callback to set the sensor occupancy if
        # sensor is known, but only if that isn't already set - it helps when
        # device notifications on triggerring sensor's activity aren't receveid
        # by a reason
        if sensor and not sensor.occupancy:
            await self.on_sensor_activity(event_id, zone_name, True)

        G90Callback.invoke(
            self._alarm_cb, event_id, zone_name, extra_data
        )

    @property
    def alarm_callback(self) -> Optional[AlarmCallback]:
        """
        The device alarm callback, which is invoked when device alarm triggers.
        """
        return self._alarm_cb

    @alarm_callback.setter
    def alarm_callback(self, value: AlarmCallback) -> None:
        self._alarm_cb = value

    async def on_low_battery(self, event_id: int, zone_name: str) -> None:
        """
        Invoked when the sensor reports on low battery. Fires
        corresponding callback if set by the user with
        :attr:`.on_low_battery_callback`.

        Please note the method is for internal use by the class.

        :param event_id: Index of the sensor triggered alarm
        :param zone_name: Sensor name
        """
        _LOGGER.debug('on_low_battery: %s %s', event_id, zone_name)
        sensor = await self.find_sensor(event_id, zone_name)
        if sensor:
            # Set the low battery flag on the sensor
            # pylint: disable=protected-access
            sensor._set_low_battery(True)
            # Invoke per-sensor callback if provided
            G90Callback.invoke(sensor.low_battery_callback)

        G90Callback.invoke(self._low_battery_cb, event_id, zone_name)

    @property
    def low_battery_callback(self) -> Optional[LowBatteryCallback]:
        """
        Low battery callback, which is invoked when sensor reports the
        condition.
        """
        return self._low_battery_cb

    @low_battery_callback.setter
    def low_battery_callback(self, value: LowBatteryCallback) -> None:
        self._low_battery_cb = value

    async def on_sos(
        self, event_id: int, zone_name: str, is_host_sos: bool
    ) -> None:
        """
        Invoked when SOS alert is triggered. Fires corresponding callback if
        set by the user with :attr:`.sos_callback`.

        Please note the method is for internal use by the class.

        :param event_id: Index of the sensor triggered alarm
        :param zone_name: Sensor name
        :param is_host_sos:
          Flag indicating if the SOS alert is triggered by the panel itself
          (host)
        """
        _LOGGER.debug('on_sos: %s %s %s', event_id, zone_name, is_host_sos)
        G90Callback.invoke(self._sos_cb, event_id, zone_name, is_host_sos)

        # Also report the event as alarm for unification, hard-coding the
        # sensor name in case of host SOS
        await self.on_alarm(event_id, 'Host SOS' if is_host_sos else zone_name)

        if not is_host_sos:
            # Also report the remote button press for SOS - the panel will not
            # send corresponding alert
            await self.on_remote_button_press(
                event_id, zone_name, G90RemoteButtonStates.SOS
            )

    @property
    def sos_callback(self) -> Optional[SosCallback]:
        """
        SOS callback, which is invoked when SOS alert is triggered.
        """
        return self._sos_cb

    @sos_callback.setter
    def sos_callback(self, value: SosCallback) -> None:
        self._sos_cb = value

    async def on_remote_button_press(
        self, event_id: int, zone_name: str, button: G90RemoteButtonStates
    ) -> None:
        """
        Invoked when remote button is pressed. Fires corresponding callback if
        set by the user with :attr:`.remote_button_press_callback`.

        Please note the method is for internal use by the class.

        :param event_id: Index of the sensor triggered alarm
        :param zone_name: Sensor name
        :param button: The button pressed
        """
        _LOGGER.debug(
            'on_remote_button_press: %s %s %s', event_id, zone_name, button
        )
        G90Callback.invoke(
            self._remote_button_press_cb, event_id, zone_name, button
        )

        # Also report the event as sensor activity for unification (remote is
        # just a special type of the sensor)
        await self.on_sensor_activity(event_id, zone_name, True)

    @property
    def remote_button_press_callback(
        self
    ) -> Optional[RemoteButtonPressCallback]:
        """
        Remote button press callback, which is invoked when remote button is
        pressed.
        """
        return self._remote_button_press_cb

    @remote_button_press_callback.setter
    def remote_button_press_callback(
        self, value: RemoteButtonPressCallback
    ) -> None:
        self._remote_button_press_cb = value

    async def listen_device_notifications(self) -> None:
        """
        Starts internal listener for device notifications/alerts.
        """
        await self.listen()

    def close_device_notifications(self) -> None:
        """
        Closes the listener for device notifications/alerts.
        """
        self.close()

    async def arm_away(self) -> None:
        """
        Arms the device in away mode.
        """
        state = G90ArmDisarmTypes.ARM_AWAY
        await self.command(G90Commands.SETHOSTSTATUS,
                           [state])

    async def arm_home(self) -> None:
        """
        Arms the device in home mode.
        """
        state = G90ArmDisarmTypes.ARM_HOME
        await self.command(G90Commands.SETHOSTSTATUS,
                           [state])

    async def disarm(self) -> None:
        """
        Disarms the device.
        """
        state = G90ArmDisarmTypes.DISARM
        await self.command(G90Commands.SETHOSTSTATUS,
                           [state])

    @property
    def sms_alert_when_armed(self) -> bool:
        """
        When enabled, allows to save costs on SMS by having corresponding alert
        enabled only when device is armed.
        """
        return self._sms_alert_when_armed

    @sms_alert_when_armed.setter
    def sms_alert_when_armed(self, value: bool) -> None:
        self._sms_alert_when_armed = value

    async def start_simulating_alerts_from_history(
        self, interval: float = 5, history_depth: int = 5
    ) -> None:
        """
        Starts the separate task to simulate device alerts from history
        entries.

        The listener for device notifications will be stopped, so device
        notifications will not be processed thus resulting in possible
        duplicated if those could be received from the network.

        :param interval: Interval (in seconds) between polling for newer
          history entities
        :param history_depth: Amount of history entries to fetch during
          each polling cycle
        """
        # Remember if device notifications listener has been started already
        self._alert_simulation_start_listener_back = self.listener_started
        # And then stop it
        self.close()

        # Start the task
        self._alert_simulation_task = asyncio.create_task(
            self._simulate_alerts_from_history(interval, history_depth)
        )

    async def stop_simulating_alerts_from_history(self) -> None:
        """
        Stops the task simulating device alerts from history entries.

        The listener for device notifications will be started back, if it was
        running when simulation has been started.
        """
        # Stop the task simulating the device alerts from history if it was
        # running
        if self._alert_simulation_task:
            self._alert_simulation_task.cancel()
            self._alert_simulation_task = None

        # Start device notifications listener back if it was running when
        # simulated alerts have been enabled
        if self._alert_simulation_start_listener_back:
            await self.listen()

    async def _simulate_alerts_from_history(
        self, interval: float, history_depth: int
    ) -> None:
        """
        Periodically fetches history entries from the device and simulates
        device alerts off of those.

        Only the history entries occur after the process is started are
        handled, to avoid triggering callbacks retrospectively.

        See :meth:`.start_simulating_alerts_from_history` for the parameters.
        """
        last_history_ts = None

        _LOGGER.debug(
            'Simulating device alerts from history:'
            ' interval %s, history depth %s',
            interval, history_depth
        )
        while True:
            try:
                # Retrieve the history entries of the specified amount - full
                # history retrieval might be an unnecessary long operation
                history = await self.history(count=history_depth)

                # Initial iteration where no timestamp of most recent history
                # entry is recorded - do that and skip to next iteration, since
                # it isn't yet known what entries would be considered as new
                # ones.
                # Empty history will skip recording the timestamp and the
                # looping over entries below, effectively skipping to next
                # iteration
                if not last_history_ts and history:
                    # First entry in the list is assumed to be the most recent
                    # one
                    last_history_ts = history[0].datetime
                    _LOGGER.debug(
                        'Initial time stamp of last history entry: %s',
                        last_history_ts
                    )
                    continue

                # Process history entries from older to newer to preserve the
                # order of happenings
                for item in reversed(history):
                    # Process only the entries newer than one been recorded as
                    # most recent one
                    if last_history_ts and item.datetime > last_history_ts:
                        _LOGGER.debug(
                            'Found newer history entry: %s, simulating alert',
                            repr(item)
                        )
                        # Send the history entry down the device notification
                        # code as alert, as if it came from the device and its
                        # notifications port
                        self._handle_alert(
                            (self._host, self._notifications_local_port),
                            item.as_device_alert(),
                            # Skip verifying device GUID, since history entry
                            # don't have it
                            verify_device_id=False
                        )

                        # Record the entry as most recent one
                        last_history_ts = item.datetime
                        _LOGGER.debug(
                            'Time stamp of last history entry: %s',
                            last_history_ts
                        )
            except (G90Error, G90TimeoutError) as exc:
                _LOGGER.debug(
                    'Error interacting with device, ignoring %s', repr(exc)
                )
            except Exception as exc:
                _LOGGER.error(
                    'Exception simulating device alerts from history: %s',
                    repr(exc)
                )
                raise exc

            # Sleep to next iteration
            await asyncio.sleep(interval)



"""
Tests for G90Alarm class
"""
import asyncio
from itertools import cycle
from unittest.mock import MagicMock
import pytest

from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.host_info import (
    G90HostInfo, G90HostInfoGsmStatus, G90HostInfoWifiStatus,
)
from pyg90alarm.entities.sensor import (
    G90Sensor,
)
from pyg90alarm.entities.device import (
    G90Device,
)

from pyg90alarm.host_status import (
    G90HostStatus,
)
from pyg90alarm.user_data_crc import (
    G90UserDataCRC,
)
from pyg90alarm.config import (
    G90AlertConfigFlags,
)
from pyg90alarm.const import (
    G90RemoteButtonStates,
)


from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[100,[3,"PHONE","PRODUCT","206","206"]]IEND\0',
])
async def test_host_status(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving host status from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    res = await g90.get_host_status()

    assert mock_device.recv_data == [b'ISTART[100,100,""]IEND\0']
    assert isinstance(res, G90HostStatus)
    assert res.host_status == 3
    assert res.product_name == 'PRODUCT'
    assert res.host_phone_number == 'PHONE'
    assert isinstance(res._asdict(), dict)


@pytest.mark.g90device(sent_data=[
    b'ISTART[206,'
    b'["DUMMYGUID","DUMMYPRODUCT",'
    b'"1.2","1.1","206","206",3,3,0,2,"4242",50,100]]IEND\0',
])
async def test_host_info(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving host information from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    res = await g90.get_host_info()

    assert mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']
    assert isinstance(res, G90HostInfo)
    assert res.host_guid == 'DUMMYGUID'
    assert res.product_name == 'DUMMYPRODUCT'
    assert res.gsm_status == G90HostInfoGsmStatus.OPERATIONAL
    assert res.wifi_status == G90HostInfoWifiStatus.OPERATIONAL
    assert isinstance(res._asdict(), dict)


@pytest.mark.g90device(sent_data=[
    b'ISTART[160,["1","0xab","3","4","5","6"]]IEND\0',
    b'ISTART[160,["1","0xab","3","4","5","6"]]IEND\0',
])
async def test_user_data_crc(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving user data CRCs from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    crc = await g90.get_user_data_crc()
    prop_crc = await g90.user_data_crc

    assert crc == prop_crc
    assert mock_device.recv_data == [
        b'ISTART[160,160,""]IEND\0',
        b'ISTART[160,160,""]IEND\0',
    ]
    assert isinstance(crc, G90UserDataCRC)
    assert crc.sensor_list == '1'
    assert crc.device_list == '0xab'
    assert crc.history_list == '3'
    assert crc.scene_list == '4'
    assert crc.ifttt_list == '5'
    assert crc.fingerprint_list == '6'


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
])
async def test_devices(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving devices from the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices

    assert devices == prop_devices
    assert mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]
    assert len(devices) == 1
    assert isinstance(devices, list)
    assert isinstance(devices[0], G90Device)
    assert devices[0].name == 'Switch'
    assert devices[0].index == 10
    assert isinstance(devices[0]._asdict(), dict)


# Provide an endless sequence of simulated panel responses for the devices
# list. Each attempt will simulate a single device. This sequence will prevent
# `G90TimeoutError` if the code under test initiates more exchanges with the
# panel than the simulated data contains.
@pytest.mark.g90device(sent_data=cycle([
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
]))
async def test_get_devices_concurrent(mock_device: DeviceMock) -> None:
    """
    Tests for concurrently retrieving list of devices produces consistent
    results.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    g90.paginated_result = MagicMock(  # type: ignore[method-assign]
        spec=g90.paginated_result, wraps=g90.paginated_result
    )

    # Issue two concurrent requests to retrieve devices
    res = await asyncio.gather(g90.get_devices(), g90.get_devices())
    # Ensure only single exchange with the panel
    g90.paginated_result.assert_called_once()
    # While `pylint` demands use of generator, the comprehension is used
    # instead for ease of trroubleshooting test failures as it will show the
    # list elements, not just generator instance
    # pylint: disable=use-a-generator
    assert all([len(x) == 1 for x in res])


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,2,0,""]]]IEND\0'
])
async def test_multinode_device(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving multi-node devices (e.g. multi-channel switch) from
    the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices

    assert devices == prop_devices
    assert mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]
    assert len(devices) == 2
    assert isinstance(devices, list)
    assert isinstance(devices[0], G90Device)
    assert devices[0].name == 'Switch#1'
    assert devices[0].index == 10
    assert devices[0].subindex == 0
    assert devices[1].name == 'Switch#2'
    assert devices[1].index == 10
    assert devices[1].subindex == 1


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
    b'ISTARTIEND\0',
    b'ISTARTIEND\0',
])
async def test_control_device(mock_device: DeviceMock) -> None:
    """
    Tests for controlling devices from the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices

    assert devices == prop_devices
    await devices[0].turn_on()
    await devices[0].turn_off()
    assert mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
        b'ISTART[137,137,[137,[10,0,0]]]IEND\0',
        b'ISTART[137,137,[137,[10,1,0]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
])
async def test_single_sensor(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving single sensor from the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors

    assert sensors == prop_sensors
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
    ]
    assert len(sensors) == 1
    assert isinstance(sensors, list)
    assert isinstance(sensors[0], G90Sensor)
    assert sensors[0].name == 'Remote'
    assert sensors[0].index == 10
    assert isinstance(sensors[0]._asdict(), dict)


# See `test_get_devices_concurrent` for the explanation of the test
@pytest.mark.g90device(sent_data=cycle([
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
]))
async def test_get_sensors_concurrent(mock_device: DeviceMock) -> None:
    """
    Tests for concurrently retrieving list of sensors produces consistent
    results.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    g90.paginated_result = MagicMock(  # type: ignore[method-assign]
        spec=g90.paginated_result, wraps=g90.paginated_result
    )

    res = await asyncio.gather(g90.get_sensors(), g90.get_sensors())
    g90.paginated_result.assert_called_once()
    # pylint: disable=use-a-generator
    assert all([len(x) == 1 for x in res])


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[3,1,3],["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Cord 1",12,0,126,1,0,32,0,5,16,1,0,""]'
    b']]IEND\0',
])
async def test_multiple_sensors_shorter_than_page(
    mock_device: DeviceMock
) -> None:
    """
    Tests for retrieving multiple sensors from the panel, while the number of
    those is shorter than a single page.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors

    assert sensors == prop_sensors
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
    ]
    assert len(sensors) == 3
    assert isinstance(sensors, list)
    assert isinstance(sensors[0], G90Sensor)
    assert sensors[0].name == 'Remote 1'
    assert sensors[0].index == 10
    assert sensors[0].is_wireless is True
    assert isinstance(sensors[1], G90Sensor)
    assert sensors[1].name == 'Remote 2'
    assert sensors[1].index == 11
    assert sensors[1].is_wireless is True
    assert isinstance(sensors[2], G90Sensor)
    assert sensors[2].name == 'Cord 1'
    assert sensors[2].index == 12
    assert sensors[2].is_wireless is False


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[11,1,10],'
    b'["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 3",12,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 4",13,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 5",14,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 6",15,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 7",16,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 8",17,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 9",18,0,10,1,0,32,0,0,16,1,0,""],'
    b'["Remote 10",19,0,10,1,0,32,0,0,16,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[11,11,1],'
    b'["Remote 11",20,0,10,1,0,32,0,0,16,1,0,""]'
    b']]IEND\0',
])
async def test_multiple_sensors_longer_than_page(
    mock_device: DeviceMock
) -> None:
    """
    Tests for retrieving multiple sensors from the panel, while the number of
    those is longer than a single page.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors

    assert sensors == prop_sensors
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[11,11]]]IEND\0',
    ]
    assert len(sensors) == 11
    assert isinstance(sensors, list)
    assert isinstance(sensors[0], G90Sensor)
    assert sensors[0].name == 'Remote 1'
    assert sensors[0].index == 10


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[170,[5,[10,"Remote"]]]\0',
    ]
)
async def test_sensor_callback(mock_device: DeviceMock) -> None:
    """
    Tests for sensor callback.
    """
    reset_interval = 0.5
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   reset_occupancy_interval=reset_interval,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors

    assert sensors == prop_sensors
    future = asyncio.get_running_loop().create_future()
    sensor = [x for x in sensors if x.index == 10 and x.name == 'Remote']
    sensor[0].state_callback = lambda *args: future.set_result(True)

    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    g90.close_device_notifications()
    # Once passed this implies the G90Alarm sensor callback works as
    # expected, as it updates the occupancy states of the sensor
    assert sensor[0].occupancy

    # Re-create future since the callback will be invoked again when the
    # occupancy state is reset
    future = asyncio.get_running_loop().create_future()
    # Verify the occupancy state is reset upon configured interval
    await asyncio.sleep(reset_interval)
    assert not sensor[0].occupancy


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",26,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,26,1,4,"Remote","DUMMYGUID",1719223959,0,[""]]]\0',
        # Simulate sensor activity, which should reset low battery state for it
        b'[170,[5,[26,"Remote"]]]\0',
    ]
)
async def test_sensor_low_battery_callback(mock_device: DeviceMock) -> None:
    """
    Tests for sensor low battery callback.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors

    assert sensors == prop_sensors
    future = asyncio.get_running_loop().create_future()
    sensor = [x for x in sensors if x.index == 26 and x.name == 'Remote']
    low_battery_sensor_cb = MagicMock()
    low_battery_sensor_cb.side_effect = lambda *args: future.set_result(True)
    sensor[0].low_battery_callback = low_battery_sensor_cb
    low_battery_cb = MagicMock()
    g90.low_battery_callback = low_battery_cb

    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    low_battery_sensor_cb.assert_called_once_with()
    low_battery_cb.assert_called_once_with(26, 'Remote')
    # Verify the low battery state is set upon receiving the notification
    assert sensor[0].is_low_battery is True

    # Signal the second notification is ready, the future has to be re-created
    # as the corresponding callback will be fired again
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    # Verify the low battery state is reset upon sensor activity
    assert sensor[0].is_low_battery is False

    g90.close_device_notifications()


@pytest.mark.g90device(
    notification_data=[
        b'[170,[1,[1]]]\0'
    ]
)
async def test_armdisarm_callback(mock_device: DeviceMock) -> None:
    """
    Tests for arm/disarm callback.
    """
    future = asyncio.get_running_loop().create_future()
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future.set_result(True)
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    g90.armdisarm_callback = armdisarm_cb
    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    g90.close_device_notifications()
    armdisarm_cb.assert_called_once_with(1)


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Door",100,0,1,1,0,32,0,0,16,1,0,""]]]IEND\0',
        # The GETNOTICEFLAG command is invoked twice, for each of
        # alerts simulated below
        b'ISTART[117,[256]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,100,1,1,"Door","DUMMYGUID",1631545189,0,[""]]]\0',
        b'[208,[4,100,1,0,"Door","DUMMYGUID",1631545189,0,[""]]]\0',
    ]
)
async def test_door_open_close_callback(mock_device: DeviceMock) -> None:
    """
    Tests for door open/close callback.
    """
    future = asyncio.get_running_loop().create_future()
    door_open_close_cb = MagicMock()
    door_open_close_cb.side_effect = lambda *args: future.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    g90.door_open_close_callback = door_open_close_cb

    # Simulate two device alerts - for opening (this one) and then closing the
    # door (see below)
    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Corresponding sensor should turn to occupied (=door opened)
    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors
    assert sensors == prop_sensors
    assert sensors[0].occupancy

    # Simulate second device alert for door close
    await mock_device.send_next_notification()
    # Signal the second alert is ready, the future has to be re-created as
    # the corresponding callback will be fired again
    future = asyncio.get_running_loop().create_future()

    await asyncio.wait([future], timeout=0.1)
    # The sensor should become inactive (=door closed)
    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors
    assert sensors == prop_sensors
    assert not sensors[0].occupancy

    g90.close_device_notifications()


@pytest.mark.g90device(
    sent_data=[
        # Simulate panel with two sensors
        b'ISTART[102,'
        b'[[2,1,2],'
        b'["Hall",100,0,1,1,0,32,0,0,16,1,0,""],'
        b'["Room",101,0,1,1,0,32,0,0,16,1,0,""]'
        b']]IEND\0',
        # Alert configuration, used by sensor activity callback invoked when
        # handling alarm
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[3,100,1,1,"Hall","DUMMYGUID",1630876128,0,[""]]]\0',
        b'[208,[3,101,1,1,"Room","DUMMYGUID",1630876128,0,[""]]]\0',
        b'[208,[3,102,1,1,"No Room","DUMMYGUID",1630876128,0,[""]]]\0',
    ]
)
async def test_alarm_callback(mock_device: DeviceMock) -> None:
    """
    Tests for alarm callback.
    """
    future = asyncio.get_running_loop().create_future()
    alarm_cb = MagicMock()
    alarm_cb.side_effect = lambda *args: future.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    sensors = await g90.get_sensors()
    # Set extra data for the 1st sensor
    sensors[0].extra_data = 'Dummy extra data'

    g90.alarm_callback = alarm_cb
    await g90.listen_device_notifications()
    # Simulate three alarm notifications - for sensor with extra data,
    # another for sensor with no extra data, and third for non-existent
    # sensor
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Verify extra data is passed to the callback
    alarm_cb.assert_called_once_with(100, 'Hall', 'Dummy extra data')
    # Verify the triggering sensor is set to active
    assert sensors[0].occupancy is True

    # Simulate alarm for sensor with no extra data
    alarm_cb.reset_mock()
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Verify no extra data is passed to the callback
    alarm_cb.assert_called_once_with(101, 'Room', None)
    # Verify the triggering sensor is set to active
    assert sensors[0].occupancy is True

    # Simulate alarm for non-existent sensor
    alarm_cb.reset_mock()
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Simulate callback is called with no data
    alarm_cb.assert_called_once_with(102, 'No Room', None)

    g90.close_device_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",11,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        # Host SOS
        b'[208,[1,1,0,0,"","DUMMYGUID",1734175050,0,[""]]]\0',
        # Remote SOS
        b'[208,[3,11,10,3,"Remote","DUMMYGUID",1734177048,0,[""]]]\0',
    ]
)
async def test_sos_callback(mock_device: DeviceMock) -> None:
    """
    Tests for SOS callback.
    """
    future_sos = asyncio.get_running_loop().create_future()
    future_alarm = asyncio.get_running_loop().create_future()

    sos_cb = MagicMock()
    sos_cb.side_effect = lambda *args: future_sos.set_result(True)
    alarm_cb = MagicMock()
    alarm_cb.side_effect = lambda *args: future_alarm.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    g90.sos_callback = sos_cb
    g90.alarm_callback = alarm_cb

    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future_sos, future_alarm], timeout=0.1)
    sos_cb.assert_called_once_with(1, 'Host SOS', True)
    alarm_cb.assert_called_once_with(1, 'Host SOS', None)

    sos_cb.reset_mock()
    alarm_cb.reset_mock()
    future_sos = asyncio.get_running_loop().create_future()
    future_alarm = asyncio.get_running_loop().create_future()
    future_button = asyncio.get_running_loop().create_future()
    button_cb = MagicMock()
    button_cb.side_effect = lambda *args: future_button.set_result(True)
    g90.remote_button_press_callback = button_cb
    await mock_device.send_next_notification()
    await asyncio.wait([future_sos, future_alarm, future_button], timeout=0.1)
    sos_cb.assert_called_once_with(11, 'Remote', False)
    alarm_cb.assert_called_once_with(11, 'Remote', None)
    # Button press callback should be called with the remote button state, but
    # only for SOS initiated by the remote
    button_cb.assert_called_once_with(11, 'Remote', G90RemoteButtonStates.SOS)

    g90.close_device_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",11,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,11,10,0,"Remote","GA18018B3001021",1734176900,0,[""]]]\0',
    ]
)
async def test_remote_button_callback(mock_device: DeviceMock) -> None:
    """
    Tests for remote button callback.
    """
    future_sensor = asyncio.get_running_loop().create_future()
    future_button = asyncio.get_running_loop().create_future()
    sensor_cb = MagicMock()
    sensor_cb.side_effect = lambda *args: future_sensor.set_result(True)
    button_cb = MagicMock()
    button_cb.side_effect = lambda *args: future_button.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    g90.sensor_callback = sensor_cb
    g90.remote_button_press_callback = button_cb

    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future_sensor, future_button], timeout=0.1)
    sensor_cb.assert_called_once_with(11, 'Remote', True)
    button_cb.assert_called_once_with(
        11, 'Remote', G90RemoteButtonStates.ARM_AWAY
    )

    g90.close_device_notifications()


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_arm_away(mock_device: DeviceMock) -> None:
    """
    Tests for arming the device in away mode.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.arm_away()
    assert mock_device.recv_data == [
        b'ISTART[101,101,[101,[1]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_arm_home(mock_device: DeviceMock) -> None:
    """
    Tests for arming the device in home mode.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.arm_home()
    assert mock_device.recv_data == [
        b'ISTART[101,101,[101,[2]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_disarm(mock_device: DeviceMock) -> None:
    """
    Tests for disarming the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.disarm()
    assert mock_device.recv_data == [
        b'ISTART[101,101,[101,[3]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[117,[1]]IEND\0',
])
async def test_alert_config(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving alert configuration from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    config = await g90.get_alert_config()
    prop_config = await g90.alert_config
    assert config == prop_config
    assert mock_device.recv_data == [b'ISTART[117,117,""]IEND\0']
    assert isinstance(config, G90AlertConfigFlags)


@pytest.mark.g90device(sent_data=[
    b"ISTART[117,[1]]IEND\0",
    b"ISTART[117,[3]]IEND\0",
    b"ISTARTIEND\0",
])
async def test_set_alert_config(mock_device: DeviceMock) -> None:
    """
    Tests for setting alert configuration to the the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    await g90.set_alert_config(
        await g90.get_alert_config()
        | G90AlertConfigFlags.AC_POWER_FAILURE  # noqa:W503
        | G90AlertConfigFlags.HOST_LOW_VOLTAGE  # noqa:W503
    )
    assert mock_device.recv_data == [
        b'ISTART[117,117,""]IEND\0',
        b'ISTART[117,117,""]IEND\0',
        b"ISTART[116,116,[116,[9]]]IEND\0",
    ]
    # Validate we retrieve same alert configuration just has been set
    assert await g90.get_alert_config() == (
        G90AlertConfigFlags.AC_POWER_FAILURE
        | G90AlertConfigFlags.HOST_LOW_VOLTAGE  # noqa:W503
    )


@pytest.mark.g90device(
    sent_data=[
        # First command to get alert configuration is from
        # `G90Alarm.get_alert_config()` property
        b"ISTART[117,[1]]IEND\0",
        # Second command for same is invoked by `G90Alarm.set_alert_config`
        # that checks if alert config has been modified externally
        b"ISTART[117,[1]]IEND\0",
        b"ISTARTIEND\0",
    ],
    notification_data=[
        b'[170,[1,[1]]]\0',
    ]
)
async def test_sms_alert_when_armed(mock_device: DeviceMock) -> None:
    """
    Tests for enabling SMS alerts when device is armed.
    """
    future = asyncio.get_running_loop().create_future()
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future.set_result(True)
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    g90.armdisarm_callback = armdisarm_cb
    g90.sms_alert_when_armed = True
    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    g90.close_device_notifications()
    assert mock_device.recv_data == [
        b'ISTART[117,117,""]IEND\0',
        b'ISTART[117,117,""]IEND\0',
        b"ISTART[116,116,[116,[513]]]IEND\0",
    ]


@pytest.mark.g90device(
    sent_data=[
        # See above for the clarification on command sequence
        b"ISTART[117,[513]]IEND\0",
        b"ISTART[117,[513]]IEND\0",
        b"ISTARTIEND\0",
    ],
    notification_data=[
        b'[170,[1,[3]]]\0',
    ]
)
async def test_sms_alert_when_disarmed(mock_device: DeviceMock) -> None:
    """
    Tests for disabling SMS alerts when device is disarmed.
    """
    future = asyncio.get_running_loop().create_future()
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future.set_result(True)
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port,
                   notifications_local_host=mock_device.notification_host,
                   notifications_local_port=mock_device.notification_port)
    g90.armdisarm_callback = armdisarm_cb
    g90.sms_alert_when_armed = True
    await g90.listen_device_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    g90.close_device_notifications()
    assert mock_device.recv_data == [
        b'ISTART[117,117,""]IEND\0',
        b'ISTART[117,117,""]IEND\0',
        b"ISTART[116,116,[116,[1]]]IEND\0",
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Night Light1",11,0,138,0,0,33,0,0,17,1,0,""],'
    b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[2,2,1],'
    b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_sensor_disable(mock_device: DeviceMock) -> None:
    """
    Tests for disabling a sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors
    assert sensors == prop_sensors
    assert sensors[1].enabled
    await sensors[1].set_enabled(False)
    assert not sensors[1].enabled
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[2,2]]]IEND\0',
        b'ISTART[103,103,[103,'
        b'["Night Light2",10,0,138,0,0,32,0,0,17,1,0,2,"060A0600"]'
        b']]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Night Light1",11,0,138,0,0,33,0,0,17,1,0,""],'
    b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,'
    b'[[2,2,1],'
    b'["Night Light2",10,0,138,0,0,1,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_sensor_disable_externally_modified(
    mock_device: DeviceMock
) -> None:
    """
    Tests for disabling a sensor that has been modified externally.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors
    assert sensors == prop_sensors
    assert sensors[1].enabled
    await sensors[1].set_enabled(False)
    assert sensors[1].enabled
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[2,2]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],["Unsupported",10,0,255,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_sensor_unsupported_disable(mock_device: DeviceMock) -> None:
    """
    Tests for disabling an unsupported sensor.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.get_sensors()
    prop_sensors = await g90.sensors
    assert sensors == prop_sensors
    assert sensors[0].enabled
    await sensors[0].set_enabled(False)
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Night Light1",11,0,138,0,0,33,0,0,17,1,0,""],'
    b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b'ISTART[102,[[2,2,0]]]IEND\0',
])
async def test_sensor_disable_sensor_not_found_on_refresh(
    mock_device: DeviceMock
) -> None:
    """
    Tests for disabling a sensor that is not found on refresh.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensors = await g90.get_sensors()
    assert sensors[1].enabled
    await sensors[1].set_enabled(False)
    assert sensors[1].enabled
    assert mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,10]]]IEND\0',
        b'ISTART[102,102,[102,[2,2]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Water",10,0,7,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_device_unsupported_disable(mock_device: DeviceMock) -> None:
    """
    Tests for disabling an unsupported device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices
    assert devices == prop_devices
    assert devices[0].enabled
    await devices[0].set_enabled(False)
    assert mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------
import sys
from sphinx.util.inspect import getdoc
from sphinx.util.docstrings import separate_metadata
sys.path.extend(['../src'])

# -- Project information -----------------------------------------------------

project = 'pyg90alarm'
copyright = '2021, Ilia Sotnikov'
author = 'Ilia Sotnikov'


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'enum_tools.autoenum',
    'sphinx.ext.autosectionlabel',
    'sphinx_autodoc_typehints',
]

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

autodoc_default_options = {
    'members': True,
    'inherited-members': True,
    'show-inheritance': True,
    'member-order': 'bysource',
    'class-doc-from': 'both',
}

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}


def autodoc_skip_class(_app, what, name, obj, _skip, _options):
    """
    Allows to declare a class as private by specifying `:meta private:` in its
    docstring.
    """
    if what == 'class':
        try:
            doc = getdoc(obj, cls=name, name=name)
            doc, metadata = separate_metadata(doc)

            if 'private' in metadata:
                return True
        except Exception:
            # Treat any exceptions in the code above non-fatal
            pass
    # Defer to other handlers to decide
    return None


def setup(app):
    """
    Registers custom handler for `autodoc-skip-member` event.
    """
    app.connect('autodoc-skip-member', autodoc_skip_class)



# Copyright (c) 2023 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Performs runtime configuration and exposes custom fixtures for Pytest.
"""
from __future__ import annotations
from typing import AsyncIterator, Callable
import pytest
from .device_mock import DeviceMock


def pytest_configure(config: pytest.Config) -> None:
    """
    Configures `pytest`.
    """
    config.addinivalue_line("markers", "g90device")


@pytest.fixture
async def mock_device(
    request: pytest.FixtureRequest, unused_udp_port_factory: Callable[..., int]
) -> AsyncIterator[DeviceMock]:
    """
    Fixture to instantiate a simulated G90 device allocating random unused
    ports for network exchanges.

    The fixture should be customized with `g90device` mark containing
    `sent_data` and `notification_data` lists of bytes to simulate the messages
    the device sends back to client and notification messages sent to client,
    respectively.
    """
    marker = getattr(
        request.node
        .get_closest_marker('g90device'),
        'kwargs', {}
    )
    data = marker.get('sent_data', [])
    notification_data = marker.get('notification_data', [])

    # Allocate unused ports to listen for client requests on, and to send
    # notification messages to, respectively

    # Note the `unised_udp_port_factory` comes from `pytest-asyncio` package
    device_port = unused_udp_port_factory()
    notification_port = unused_udp_port_factory()
    device = DeviceMock(
        data, notification_data,
        device_port=device_port, notification_port=notification_port
    )
    await device.start()
    yield device
    device.stop()



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides interface to devices (switches) of G90 alarm panel.
"""
from __future__ import annotations
import logging
from .sensor import G90Sensor
from ..const import G90Commands


_LOGGER = logging.getLogger(__name__)


class G90Device(G90Sensor):
    """
    Interacts with device (relay) on G90 alarm panel.
    """

    async def turn_on(self) -> None:
        """
        Turns on the device (relay)
        """
        await self.parent.command(G90Commands.CONTROLDEVICE,
                                  [self.index, 0, self.subindex])

    async def turn_off(self) -> None:
        """
        Turns off the device (relay)
        """
        await self.parent.command(G90Commands.CONTROLDEVICE,
                                  [self.index, 1, self.subindex])

    @property
    def supports_enable_disable(self) -> bool:
        """
        Indicates if disabling/enabling the device (relay) is supported.

        :return: Support for enabling/disabling the device
        """
        # No support for manipulating of disable/enabled for the device, since
        # single protocol entity read from the G90 alarm panel results in
        # multiple `G90Device` instances and changing the state would
        # subsequently require a design change to allow multiple entities to
        # reflect that. Multiple device entities are for multi-channel relays
        # mostly.
        return False

    async def set_enabled(self, value: bool) -> None:
        """
        Changes the disabled/enabled state of the device (relay).

        :param value: Whether to enable or disable the device
        """
        _LOGGER.warning(
            'Manipulating with enable/disable for device is unsupported'
        )



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements support for notifications/alerts sent by G90 alarm panel.
"""
import json
import logging
from typing import (
    Optional, Tuple, List, Any
)
from dataclasses import dataclass
import asyncio
from asyncio.transports import BaseTransport
from asyncio.protocols import DatagramProtocol
from .callback import G90Callback
from .const import (
    G90MessageTypes,
    G90NotificationTypes,
    G90AlertTypes,
    G90AlertStateChangeTypes,
    G90ArmDisarmTypes,
    G90AlertSources,
    G90AlertStates,
    G90RemoteButtonStates,
)

_LOGGER = logging.getLogger(__name__)


@dataclass
class G90Message:
    """
    Represents the message received from the device.

    :meta private:
    """
    code: G90MessageTypes
    data: List[Any]


@dataclass
class G90Notification:
    """
    Represents the notification received from the device.

    :meta private:
    """
    kind: G90NotificationTypes
    data: List[Any]


@dataclass
class G90ZoneInfo:
    """
    Represents zone details received from the device.

    :meta private:
    """
    idx: int
    name: str


@dataclass
class G90ArmDisarmInfo:
    """
    Represents the arm/disarm state received from the device.

    :meta private:
    """
    state: int


@dataclass
class G90DeviceAlert:  # pylint: disable=too-many-instance-attributes
    """
    Represents alert received from the device.
    """
    type: G90AlertTypes
    event_id: G90AlertStateChangeTypes
    source: G90AlertSources
    state: int
    zone_name: str
    device_id: str
    unix_time: int
    resv4: int
    other: str


class G90DeviceNotifications(DatagramProtocol):
    """
    Implements support for notifications/alerts sent by alarm panel.

    There is a basic check to ensure only notifications/alerts from the correct
    device are processed - the check uses the host and port of the device, and
    the device ID (GUID) that is set by the ancestor class that implements the
    commands (e.g. :class:`G90Alarm`). The latter to work correctly needs a
    command to be performed first, one that fetches device GUID and then stores
    it using :attr:`.device_id` (e.g. :meth:`G90Alarm.get_host_info`).
    """
    def __init__(self, local_port: int, local_host: str):
        # pylint: disable=too-many-arguments
        self._notification_transport: Optional[BaseTransport] = None
        self._notifications_local_host = local_host
        self._notifications_local_port = local_port
        # Host/port of the device is configured to communicating via commands.
        # Inteded to validate if notifications/alert are received from the
        # correct device.
        self._host: Optional[str] = None
        self._port: Optional[int] = None
        # Same but for device ID (GUID) - the notifications logic uses it to
        # perform validation, but doesn't set it from messages received (it
        # will diminish the purpose of the validation, should be done by an
        # ancestor class).
        self._device_id: Optional[str] = None

    def _handle_notification(
        self, addr: Tuple[str, int], notification: G90Notification
    ) -> None:
        # Sensor activity notification
        if notification.kind == G90NotificationTypes.SENSOR_ACTIVITY:
            g90_zone_info = G90ZoneInfo(*notification.data)

            _LOGGER.debug('Sensor notification: %s', g90_zone_info)
            G90Callback.invoke(
                self.on_sensor_activity,
                g90_zone_info.idx, g90_zone_info.name
            )

            return

        # Arm/disarm notification
        if notification.kind == G90NotificationTypes.ARM_DISARM:
            g90_armdisarm_info = G90ArmDisarmInfo(
                *notification.data)
            # Map the state received from the device to corresponding enum
            state = G90ArmDisarmTypes(g90_armdisarm_info.state)

            _LOGGER.debug('Arm/disarm notification: %s',
                          state)
            G90Callback.invoke(self.on_armdisarm, state)

            return

        _LOGGER.warning('Unknown notification received from %s:%s:'
                        ' kind %s, data %s',
                        addr[0], addr[1], notification.kind, notification.data)

    def _handle_alert_sensor_activity(self, alert: G90DeviceAlert) -> bool:
        """
        Handles sensor activity alert.
        """
        if alert.source == G90AlertSources.REMOTE:
            _LOGGER.debug('Remote button press alert: %s', alert)
            G90Callback.invoke(
                self.on_remote_button_press,
                alert.event_id, alert.zone_name,
                G90RemoteButtonStates(alert.state)
            )

            return True

        if alert.state in (
            G90AlertStates.DOOR_OPEN, G90AlertStates.DOOR_CLOSE
        ):
            is_open = (
                alert.source == G90AlertSources.SENSOR
                and alert.state == G90AlertStates.DOOR_OPEN  # noqa: W503
            ) or alert.source == G90AlertSources.DOORBELL

            _LOGGER.debug('Door open_close alert: %s', alert)
            G90Callback.invoke(
                self.on_door_open_close,
                alert.event_id, alert.zone_name, is_open
            )

            return True

        if (
            alert.source == G90AlertSources.SENSOR
            and alert.state == G90AlertStates.LOW_BATTERY  # noqa: W503
        ):
            _LOGGER.debug('Low battery alert: %s', alert)
            G90Callback.invoke(
                self.on_low_battery,
                alert.event_id, alert.zone_name
            )

            return True

        return False

    def _handle_alert(
        self, addr: Tuple[str, int], alert: G90DeviceAlert,
        verify_device_id: bool = True
    ) -> None:
        handled = False

        # Stop processing when alert is received from the device with different
        # GUID (if enabled)
        if (
            verify_device_id and self.device_id
            and alert.device_id != self.device_id
        ):
            _LOGGER.error(
                "Received alert from wrong device: expected '%s', got '%s'",
                self.device_id, alert.device_id
            )
            return

        if alert.type == G90AlertTypes.SENSOR_ACTIVITY:
            handled = self._handle_alert_sensor_activity(alert)

        if alert.type == G90AlertTypes.STATE_CHANGE:
            # Define the mapping between device state received in the alert, to
            # common `G90ArmDisarmTypes` enum that is used when setting device
            # arm state and received in the corresponding notifications. The
            # primary reason is to unify state as passed down to the callbacks.
            # The map covers only subset of state changes pertinent to
            # arm/disarm state changes
            alarm_arm_disarm_state_map = {
                G90AlertStateChangeTypes.ARM_HOME: G90ArmDisarmTypes.ARM_HOME,
                G90AlertStateChangeTypes.ARM_AWAY: G90ArmDisarmTypes.ARM_AWAY,
                G90AlertStateChangeTypes.DISARM: G90ArmDisarmTypes.DISARM
            }

            state = alarm_arm_disarm_state_map.get(alert.event_id, None)
            if state:
                # We received the device state change related to arm/disarm,
                # invoke the corresponding callback
                _LOGGER.debug('Arm/disarm state change: %s', state)
                G90Callback.invoke(self.on_armdisarm, state)

            handled = True

        if alert.type == G90AlertTypes.ALARM:
            # Remote SOS
            if alert.source == G90AlertSources.REMOTE:
                _LOGGER.debug('SOS: %s', alert.zone_name)
                G90Callback.invoke(
                    self.on_sos, alert.event_id, alert.zone_name, False
                )
            # Regular alarm
            else:
                _LOGGER.debug('Alarm: %s', alert.zone_name)
                G90Callback.invoke(
                    self.on_alarm,
                    alert.event_id, alert.zone_name
                )
            handled = True

        # Host SOS
        if alert.type == G90AlertTypes.HOST_SOS:
            zone_name = 'Host SOS'

            _LOGGER.debug('SOS: Host')
            G90Callback.invoke(
                self.on_sos, alert.event_id, zone_name, True
            )

            handled = True

        if not handled:
            _LOGGER.warning('Unknown alert received from %s:%s:'
                            ' type %s, data %s',
                            addr[0], addr[1], alert.type, alert)

    # Implementation of datagram protocol,
    # https://docs.python.org/3/library/asyncio-protocol.html#datagram-protocols
    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection from the device is made.
        """

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Same but when the connection is lost.
        """

    def datagram_received(  # pylint:disable=R0911
        self, data: bytes, addr: Tuple[str, int]
    ) -> None:
        """
        Invoked when datagram is received from the device.
        """
        if self._host and self._host != addr[0]:
            _LOGGER.error(
                "Received notification/alert from wrong host '%s',"
                " expected from '%s'",
                addr[0], self._host
            )
            return

        try:
            s_data = data.decode('utf-8')
        except UnicodeDecodeError:
            _LOGGER.error('Unable to decode device message from UTF-8')
            return

        if not s_data.endswith('\0'):
            _LOGGER.error('Missing end marker in data')
            return
        payload = s_data[:-1]
        _LOGGER.debug('Received device message from %s:%s: %s',
                      addr[0], addr[1], payload)
        try:
            message = json.loads(payload)
            g90_message = G90Message(*message)
        except json.JSONDecodeError as exc:
            _LOGGER.error("Unable to parse device message '%s' as JSON: %s",
                          payload, exc)
            return
        except TypeError as exc:
            _LOGGER.error("Device message '%s' is malformed: %s",
                          payload, exc)
            return

        # Device notifications
        if g90_message.code == G90MessageTypes.NOTIFICATION:
            try:
                notification_data = G90Notification(*g90_message.data)
            except TypeError as exc:
                _LOGGER.error('Bad notification received from %s:%s: %s',
                              addr[0], addr[1], exc)
                return
            self._handle_notification(addr, notification_data)
            return

        # Device alerts
        if g90_message.code == G90MessageTypes.ALERT:
            try:
                alert_data = G90DeviceAlert(*g90_message.data)
            except TypeError as exc:
                _LOGGER.error('Bad alert received from %s:%s: %s',
                              addr[0], addr[1], exc)
                return
            self._handle_alert(addr, alert_data)
            return

        _LOGGER.warning('Unknown message received from %s:%s: %s',
                        addr[0], addr[1], message)

    async def on_armdisarm(self, state: G90ArmDisarmTypes) -> None:
        """
        Invoked when device is armed or disarmed.

        :param state: State of the device
        """

    async def on_sensor_activity(self, idx: int, name: str) -> None:
        """
        Invoked on sensor activity.

        :param idx: Index of the sensor.
        :param name: Name of the sensor.
        """

    async def on_door_open_close(
        self, event_id: int, zone_name: str, is_open: bool
    ) -> None:
        """
        Invoked when door sensor reports it opened or closed.

        :param event_id: Index of the sensor reporting the event.
        :param zone_name: Name of the sensor that reports door open/close.
        :param is_open: Indicates if the door is open.
        """

    async def on_low_battery(self, event_id: int, zone_name: str) -> None:
        """
        Invoked when a sensor reports it is low on battery.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the sensor that reports low battery.
        """

    async def on_alarm(self, event_id: int, zone_name: str) -> None:
        """
        Invoked when device triggers the alarm.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the zone that triggered the alarm.
        """

    async def on_remote_button_press(
        self, event_id: int, zone_name: str, button: G90RemoteButtonStates
    ) -> None:
        """
        Invoked when a remote button is pressed.

        Please note there will only be call to the method w/o invoking
        :meth:`G90DeviceNotifications.on_sensor_activity`.

        :param event_id: Index of the sensor associated with the remote.
        :param zone_name: Name of the sensor that reports remote button press.
        :param button: The button pressed on the remote
        """

    async def on_sos(
        self, event_id: int, zone_name: str, is_host_sos: bool
    ) -> None:
        """
        Invoked when SOS is triggered.

        Please note that the panel might not set its status to alarm
        internally, so that :meth:`G90DeviceNotifications` might need an
        explicit call in the derived class to simulate that.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the sensor that reports SOS.
        :param is_host_sos: Indicates if the SOS is host-initiated.
        """

    async def listen(self) -> None:
        """
        Listens for notifications/alers from the device.
        """
        try:
            loop = asyncio.get_running_loop()
        except AttributeError:
            loop = asyncio.get_event_loop()

        _LOGGER.debug('Creating UDP endpoint for %s:%s',
                      self._notifications_local_host,
                      self._notifications_local_port)
        (self._notification_transport,
         _protocol) = await loop.create_datagram_endpoint(
            lambda: self,
            local_addr=(
                self._notifications_local_host, self._notifications_local_port
            ))

    @property
    def listener_started(self) -> bool:
        """
        Indicates if the listener of the device notifications has been started.
        """
        return self._notification_transport is not None

    def close(self) -> None:
        """
        Closes the listener.
        """
        if self._notification_transport:
            _LOGGER.debug('No longer listening for device notifications')
            self._notification_transport.close()
            self._notification_transport = None

    @property
    def device_id(self) -> Optional[str]:
        """
        The ID (GUID) of the panel being communicated with thru commands.

        Available when any panel command receives it from the device
        (:meth:`G90Alarm.get_host_info` currently).
        """
        return self._device_id

    @device_id.setter
    def device_id(self, device_id: str) -> None:
        # Under not yet identified circumstances the device ID might be empty
        # string provided by :meth:`G90Alarm.get_host_info` - disallow that
        if not device_id or len(device_id.strip()) == 0:
            _LOGGER.debug(
                'Device ID is empty or contains whitespace only, not setting'
            )
            return

        self._device_id = device_id



# Copyright (c) 2023 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Simulates a G90 device with real network exhanges for tests.
"""
from __future__ import annotations
from typing import Optional, Tuple, List, Any, cast, Iterator
import asyncio
from asyncio.protocols import DatagramProtocol
from asyncio.transports import DatagramTransport, BaseTransport
from asyncio import Future
import logging

_LOGGER = logging.getLogger('Mock device')


class MockDeviceProtocol(DatagramProtocol):
    """
    asyncio protocol for simulate G90 device exchange over network.

    :param device_sent_data: List of datagram payloads to simulate being sent
     from device as response to client requests
    """
    def __init__(self, device_sent_data: List[bytes]):
        self._device_sent_data: Iterator[bytes] = iter(device_sent_data or [])
        self._device_recv_data: List[bytes] = []
        self._transport: Optional[DatagramTransport] = None

    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection is made.

        :param transport: asyncio transport instance
        """
        self._transport = cast(DatagramTransport, transport)

    def connection_lost(self, _err: Optional[Exception]) -> None:
        """
        Invoked when connection is lost.

        :param _err: Exception object
        """

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """
        Invoked when a datagram is received.
        """
        message = data.decode()
        self._device_recv_data.append(data)
        _LOGGER.debug('Received %s from %s:%s', message, *addr)
        try:
            sent_data = next(self._device_sent_data)
        except StopIteration:
            _LOGGER.info(
                'No more data to send, the client will experience a timeout'
                ' condition'
            )
            return

        if not self._transport:
            _LOGGER.error('Transport is not available, not sending')
            return

        _LOGGER.debug('Sent %s to %s:%s', sent_data, *addr)
        self._transport.sendto(sent_data, addr)

    @property
    def device_recv_data(self) -> List[Any]:
        """
        Returns all data received by the simulated device from the client.

        :return: List of datagram paylods received
        """
        return self._device_recv_data


class MockNotificationProtocol(DatagramProtocol):
    """
    asyncio protocol for simulate G90 device notifications over network.

    :param list(bytes) notification_data: List of datagram payloads to simulate
     being sent from device to client
    """
    def __init__(self, notification_data: bytes):
        self._notification_data = notification_data
        self._transport = None
        self._done = asyncio.get_running_loop().create_future()

    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection is made to the client and simulated
        notification is ready to be sent.

        :param transport: asyncio transport instance
        """

        remote_addr = transport.get_extra_info('peername') or (None, None)
        _LOGGER.debug(
            'Sent notification data %s to %s:%s',
            self._notification_data, *remote_addr
        )
        cast(DatagramTransport, transport).sendto(self._notification_data)
        self._done.set_result(True)

    def connection_lost(self, _err: Optional[Exception]) -> None:
        """
        Invoked when connection is lost.

        :param _err: Exception object
        """

    @property
    def is_done(self) -> Future[bool]:
        """
        Indicates if sending the notification payload has been completed.

        :return: Completed future if sending has been completed
        """
        return self._done


class DeviceMock:  # pylint:disable=too-many-instance-attributes
    """
    Simulates G90 responses and notification messages over a real network
    connection.

    :param data: List of datagram payloads to simulate being sent
     from device as response to client requests
    :param notification_data: List of datagram payloads to simulate
     notifications being sent from device to client
    :param device_port: The port simulated device listens on for client
     requests
    :param notification_port: The destination port on the client the
     notifications will be sent to
    :param device_host: The host the simulated device listen on for client
     requests
    :param notification_host: The destination client host the notifications
     will be sent to
    """
    def __init__(  # pylint:disable=too-many-arguments
        self, data: List[bytes], notification_data: List[bytes],
        device_port: int, notification_port: int,
        device_host: str = '127.0.0.1',
        notification_host: str = '127.0.0.1',
    ):
        self._host = device_host
        self._port = device_port
        self._data = data
        self._transport: Optional[BaseTransport] = None
        self._protocol: Optional[MockDeviceProtocol] = None
        self._notification_data = iter(notification_data or [])
        self._notification_port = notification_port
        self._notification_host = notification_host
        _LOGGER.debug(
            'Ports - device: %s, notification: %s',
            self._port, self._notification_port
        )

    async def start(self) -> None:
        """
        Starts listening the simulate device for client requests.
        """
        loop = asyncio.get_running_loop()
        _LOGGER.debug(
            'Creating UDP server endpoint: %s:%s', self._host, self._port
        )
        self._transport, self._protocol = await loop.create_datagram_endpoint(
            lambda: MockDeviceProtocol(self._data),
            local_addr=(self._host, self._port)
        )

    @property
    def host(self) -> str:
        """
        Returns the host the simulated device listens on.

        :return: Host name or address
        """
        return self._host

    @property
    def port(self) -> int:
        """
        Returns the port the simulated device listens on.

        :return: Port number
        """
        return self._port

    @property
    def notification_host(self) -> str:
        """
        Returns the host the simulated notifications will be sent to.

        :return: Host name or address
        """
        return self._notification_host

    @property
    def notification_port(self) -> int:
        """
        Returns the destination port will be used to send notifications to the
        client.

        :return: Port number
        """
        return self._notification_port

    @property
    def recv_data(self) -> List[bytes]:
        """
        Returns the data received by the simulated device from the client.

        :return: Data received
        """
        if not self._protocol:
            return []

        return self._protocol.device_recv_data

    def stop(self) -> None:
        """
        Stops listening for client requests.
        """
        _LOGGER.debug(
            'Closing UDP server endpoint %s:%s', self._host, self._port
        )
        if self._transport:
            self._transport.close()

    async def send_next_notification(self) -> None:
        """
        Sends next simulated notification message to the client.

        The code uses `asyncio` intentionally to cooperate with async tests.
        """
        data = None
        try:
            data = next(self._notification_data)
        except StopIteration:
            _LOGGER.info(
                'No more notification data to send, skipping'
            )
            return

        _LOGGER.debug(
            'Creating UDP notification client to %s:%s',
            self._notification_host,
            self._notification_port
        )

        loop = asyncio.get_running_loop()
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: MockNotificationProtocol(data),
            remote_addr=(self._notification_host, self._notification_port)
        )
        await asyncio.wait([protocol.is_done])
        _LOGGER.debug('Closing UDP notification client')
        transport.close()
